/**
 * Copyright (C) 2007 Fabian Greif <fabian.greif@rwth-aachen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include <util/delay.h>
#include <stdint.h>
#include <stdio.h>
#include "bool.h"

#include "defaults.h"
#include "uart.h"
#include "eeprom_functions.h"
#include "scanner.h"
#include "servo.h"
#include "io.h"

// ----------------------------------------------------------------------------
// check AVR-type

#if !defined (__AVR_ATmega48__) && !defined (__AVR_ATmega88__) && !defined (__AVR_ATmega168__)
	#error	the choosen AVR type is not supported yet!
#endif

// ----------------------------------------------------------------------------
typedef struct {
	tIOValues		io[8];
	tServoValues	servo[8];
	uint16_t		baud;
} tCommandValues;

tCommandValues command_values;

// these values will be loaded at startup if there are no other values 
// stored in the eeprom
tCommandValues command_values_defaults PROGMEM = {
	.io = {
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
		{ .periphal = 0, .output = 0, .value = 0 },
	},
	
	.servo = {
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 },
		{ .max = SIG_2_2MS, .min = SIG_0_8MS, .offset = 0, .position = 0 }
	},
	
	.baud = UART_BAUD_SELECT( 9600UL, F_CPU ),
};

// ----------------------------------------------------------------------------
#include "wrapper.h"		// TODO

// ----------------------------------------------------------------------------
tCommandDefinition PROGMEM command_table[] = {	// # = 0x23
	{ 'S', 2, servo_set_position_wrapper },		// = 0x53
	{ 's', 1, servo_get_position_wrapper },		// = 0x73
	{ 'D', 2, servo_set_offset_wrapper },		// = 0x44
	{ 'L', 5, servo_set_limits_wrapper },		// = 0x4C
	
	{ 'O', 2, set_pins_wrapper },				// = 0x4F
	{ 'I', 1, get_pins_wrapper },				// = 0x49
//	{ 'o', 4, set_timed_pins_wrapper },			// = 0x6F
	
	{ 'V', 0, get_version },					// = 0x56
	{ 'B', 1, set_baudrate },					// = 0x42
	{ 'E', 0, save_values_to_eeprom_wrapper },	// = 0x45
	{ 'd', 0, load_default_values_wrapper },	// = 0x64
	{ 'R', 0, reset_wrapper },					// = 0x52
};

// ----------------------------------------------------------------------------
static int putchar__(char c, FILE *stream) {
	uart_putc(c);
	return 0;
}

static FILE mystdout = FDEV_SETUP_STREAM(putchar__, 0, _FDEV_SETUP_WRITE);

// ----------------------------------------------------------------------------
uint8_t mcusr_mirror __attribute__ ((section (".noinit")));

void get_mcusr(void) \
	__attribute__((naked)) \
	__attribute__((section(".init3")));
void get_mcusr(void)
{
	mcusr_mirror = MCUSR;
	MCUSR = 0;
	
	wdt_disable();
}

// ----------------------------------------------------------------------------
// main loop

int main(void)
{
	DDRC |= (1<<PC0);
	PORTC |= (1<<PC0);		// Status LED on
	
	_delay_ms(10);
	_delay_ms(10);
	
	PORTC &= ~(1<<PC0);		// LED off
	
	// try to load values form eeprom, otherwise use the specified default values
	if (!eeprom_load_values((uint8_t *) &command_values, sizeof(command_values))) {
		load_default_values();
	}
	
	// enable uart communication
	uart_init( command_values.baud );
	
	// redirect stdout to the uart => now we can use printf()
	stdout = &mystdout;
	
	// enable interrupts
	sei();
	
	DEBUG("initilizing ...\n");
	DEBUG("* load io-values\n");
	io_init( command_values.io );
	
	DEBUG("* activate servos\n");
	servo_init( command_values.servo );
	
	DEBUG("* initilize command scanner\n");
	scanner_init(command_table, (sizeof(command_table) / sizeof(tCommandDefinition)));
	
	DEBUG("s8io8 ready.\n");
	
	while (1)
	{
		// try to read a character form the UART
		uint16_t tmp = uart_getc();
		
		// check if we correctly received a character
		if ((tmp & 0xff00) == 0)
		{
			// parse the character
			scanner_scan( tmp & 0xff );
		}
	}
	
	return 0;
}
