// ----------------------------------------------------------------------------
/*
 * Copyright (C) 2007 Fabian Greif <fabian.greif@rwth-aachen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
// ----------------------------------------------------------------------------
 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "defaults.h"
#include "servo.h"

// ----------------------------------------------------------------------------

tServoValues *servo_values;
uint16_t timer_values[NUM_SERVOS];

// ----------------------------------------------------------------------------
ISR(SIG_OVERFLOW0)
{
	static uint8_t servo_num = 0;
	
	TCNT0 = TIMER0_RELOAD;
	
	servo_num++;
	if (servo_num > 7)
		servo_num = 0;
	
	// select servo
	switch(servo_num)
	{
		case 0:
			PORTD |= (1<<PD3);
			break;
		case 1:
			PORTD |= (1<<PD4);
			break;
		case 2:
			PORTD |= (1<<PD5);
			break;
		case 3:
			PORTD |= (1<<PD6);
			break;
		case 4:
			PORTB |= (1<<PB2);
			break;
		case 5:
			PORTD |= (1<<PD7);
			break;
		case 6:
			PORTB |= (1<<PB1);
			break;
		case 7:
			PORTB |= (1<<PB0);
			break;
	}
	
	// start Timer 1, Prescaler = 1
	TCNT1 = timer_values[servo_num];
	TCCR1B = (1<<CS10);
}

// ----------------------------------------------------------------------------
ISR(SIG_OVERFLOW1)
{
	// clear servo-pins
	PORTB &= ~((1<<PB0)|(1<<PB1)|(1<<PB2));
	PORTD &= ~((1<<PD3)|(1<<PD4)|(1<<PD5)|(1<<PD6)|(1<<PD7));
	
	// stop timer1
	TCCR1B = 0;
}

// ----------------------------------------------------------------------------
void servo_init(tServoValues val[])
{
	servo_values = val;
	
	// update timer_values
	for (uint8_t i=0;i < NUM_SERVOS; i++) {
		servo_set_position( i, servo_values[i].position );
	}
	
	DDRB |= (1<<PB0)|(1<<PB1)|(1<<PB2);
	DDRD |= (1<<PD3)|(1<<PD4)|(1<<PD5)|(1<<PD6)|(1<<PD7);
	
	// Timer0, Prescaler = 256
	// overflow after 2,5 ms
	TCCR0B = (1 << CS02);
	TCNT0 = TIMER0_RELOAD;
	
	// enable interrupts
	TIMSK1 = (1 << TOIE1);
	TIMSK0 = (1 << TOIE0);
}

// ----------------------------------------------------------------------------
void servo_set_position(uint8_t servo, int8_t pos)
{
	if (servo >= NUM_SERVOS)
		return;
	
	int8_t offset = servo_values[servo].offset;
	int16_t position = offset + pos;
	
	// check limits
	if (position > INT8_MAX) {
		position = INT8_MAX;
	}
	else if (position < INT8_MIN) {
		position = INT8_MIN;
	}
	
	// save corrected position
	servo_values[servo].position = position - offset;
	
	// calculate timer preload value
	uint16_t max = servo_values[servo].max;
	uint16_t min = servo_values[servo].min;
	
	uint8_t upos = position + 128;
	
	timer_values[servo] = (upos * ((min - max) / 255)) + max;
	
	DEBUG("servo %d: %d (%d)\n", servo, (position - offset), offset);
}

// ----------------------------------------------------------------------------
int8_t servo_get_position(uint8_t servo)
{
	if (servo >= NUM_SERVOS)
		return 0;
	
	return (servo_values[servo].position);
}

// ----------------------------------------------------------------------------
void servo_set_offset(uint8_t servo, int8_t offset)
{
	if (servo >= NUM_SERVOS)
		return;
	
	servo_values[servo].offset = offset;
	
	// update position
	servo_set_position( servo, servo_values[servo].position );
}

	
// ----------------------------------------------------------------------------
void servo_set_limits(uint8_t servo, uint16_t max, uint16_t min)
{
	if (servo >= NUM_SERVOS)
		return;
	
	servo_values[servo].max = max;
	servo_values[servo].min = min;
}
